(**********************************************)
(*                                            *)
(*    Delphi TTblShape component v. 1.0       *)
(*             = freeware =                   *)
(*                                            *)
(*   (C) Copyright 1999, Dmitry H. Bubniewski *)
(*                                            *)
(*  e-mail: sb3@id.ru, super_bubba@mail.ru    *)
(*            ICQ # 37115452                  *)
(*                                            *)
(**********************************************)

unit TabShp;

interface

uses Windows, Messages, SysUtils, Classes, Graphics, Controls, ExtCtrls;

type
  TSizeList = class (TPersistent)
  private
    FList:TList;
    function GetCount: Integer;
    function GetItem(Index: integer): Integer;
    procedure SetItem(Index: integer; const Value: Integer);
  public
    constructor Create;
    destructor Destroy;override;
    function Add(Value:Integer):Integer;virtual;
    function SumTotal:Integer;
    procedure AddSizes(List:TSizeList);virtual;
    procedure Assign(Source:Tpersistent);override;
    procedure Clear;virtual;
    function Last:Integer;
    procedure Delete(Index:Integer);virtual;
    property Count:Integer read GetCount;
    property Items[Index:integer]:Integer read GetItem write SetItem;default;
  end;

  TTabFrameKind = (tfLeft, tfTop, tfRight, tfBottom);
  TTabFrame = set of TTabFrameKind;

  TTblShape = class(TShape)
  private
    FRow: Integer;
    FCol: Integer;
    FDownX, FDownY: Integer;
    FColResize, FRowResize:Integer;
    FHSizing, FVSizing:Boolean;
    FColSizes:TSizeList;
    FRowSizes:TSizeList;
    FTabFrame:TTabFrame;
    FTransparent: Boolean;
    FCanResize: Boolean;
    procedure SetCol(const Value: Integer);
    procedure SetRow(const Value: Integer);
    procedure CalcSplitSize;
    procedure CalcRowSize;
    function ColOffSet(Index:Integer):Integer;
    function RowOffSet(Index:Integer):Integer;
    procedure SetColSizes(const Value: TSizeList);
    procedure SetTabFrame(const Value: TTabFrame);
    procedure SetTransparent(const Value: Boolean);
    procedure SetRowSizes(const Value: TSizeList);
    procedure SetCanResize(const Value: Boolean);
  protected
    procedure DrawLineH(X:Integer);
    procedure DrawLineV(Y:Integer);
    procedure Paint; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer);override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);override;
  public
    constructor Create(AOwner:TComponent);override;
  published
    property TabFrame: TTabFrame read FTabFrame write SetTabFrame
      default [tfLeft, tfTop, tfRight, tfBottom];
    property Transparent:Boolean read FTransparent write SetTransparent;
    property TabResize:Boolean read FCanResize write SetCanResize;
    property ColSizes:TSizeList read FColSizes write SetColSizes;
    property RowSizes:TSizeList read FRowSizes write SetRowSizes;
    property Row:Integer read FRow write SetRow default 2;
    property Col:Integer read FCol write SetCol default 1;
  end;

  procedure Register;

implementation

{$R *.RES}

procedure Register;
begin
  RegisterComponents('Additional',[TTblShape]);
end;

{ TSizeList }

function TSizeList.Add(Value: Integer): Integer;
var P:PInteger;
begin
  New(P);
  P^:=Value;
  Result:=FList.Add(P);
end;

procedure TSizeList.AddSizes(List: TSizeList);
var i:integer;
begin
  for i:=0 to Pred(List.Count) do Add(List[i]);
end;

procedure TSizeList.Assign(Source: TPersistent);
begin
  if Source is TSizeList then
    begin
      Clear;
      AddSizes(TSizeList(Source));
    end
  else
    inherited Assign(Source);
end;

procedure TSizeList.Clear;
var i:integer;
begin
  for i:=0 to Pred(FList.Count) do
    Dispose(PInteger(FList.Items[i]));
  FList.Clear;
end;

constructor TSizeList.Create;
begin
  inherited Create;
  FList:=TList.Create;
end;

procedure TSizeList.Delete(Index: Integer);
begin
  Dispose(PInteger(FList.Items[Index]));
  FList.Delete(Index);
end;

destructor TSizeList.Destroy;
begin
  Clear;
  FList.Free;
  inherited Destroy;
end;

function TSizeList.GetCount: Integer;
begin
  Result:=FList.Count;
end;

function TSizeList.GetItem(Index: integer): Integer;
begin
  Result:=PInteger(FList.Items[Index])^;
end;

function TSizeList.Last: Integer;
begin
  Result:=PInteger(FList.Items[Count-1])^;;
end;

procedure TSizeList.SetItem(Index: integer;
  const Value: Integer);
begin
  PInteger(FList.Items[Index])^:=Value;
end;

function TSizeList.SumTotal: Integer;
var i:integer;
begin
  Result:=0;
  for i:=0 to Count-1 do
    Result:=Result+Items[i];
end;

{ TTblShape }

procedure TTblShape.CalcRowSize;
var i:integer;
begin
  FRowSizes.Clear;
  for i:=0 to pred(FRow) do FRowSizes.Add(Height div FRow);
end;

procedure TTblShape.CalcSplitSize;
var i:integer;
begin
  FColSizes.Clear;
  for i:=0 to pred(FCol) do
    FColSizes.Add(Width div FCol);
end;

function TTblShape.ColOffSet(Index: Integer): Integer;
var i:integer;
begin
  Result:=0;
  for i:=0 to pred(Index) do
    Result:=Result+FColSizes.Items[i];
end;

constructor TTblShape.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FRow:=2;
  FCol:=1;
  FVSizing:=False;
  FHSizing:=False;
  FCanResize:=True;
  FColSizes:=TSizeList.Create;
  FRowSizes:=TSizeList.Create;
  FTransparent:=False;
  FTabFrame:=[tfLeft, tfTop, tfRight, tfBottom];
  CalcSplitSize;
  CalcRowSize;
end;

procedure TTblShape.DrawLineH(X:Integer);
begin
   Repaint;
   with Canvas do
     begin
       Pen.Color:=clSilver;
       Pen.Width:=2;
       MoveTo(X,0);
       LineTo(X,Height);
     end;
end;

procedure TTblShape.DrawLineV(Y:Integer);
begin
   Repaint;
   with Canvas do
     begin
       Pen.Color:=clSilver;
       Pen.Width:=2;
       MoveTo(0,Y);
       LineTo(Width,Y);
     end;
end;

procedure TTblShape.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  FDownX:=X;
  FDownY:=Y;
  inherited MouseDown(Button,Shift,X,Y);
end;

procedure TTblShape.MouseMove(Shift: TShiftState; X, Y: Integer);
var i :integer;
begin
  if FCanResize then
    begin
      if ssLeft in Shift then
        begin
          if FHSizing then DrawLineH(X);
          if FVSizing then DrawLineV(Y);
        end
      else
        begin
          for i:=1 to Pred(FRow) do
            begin
              if (Y>=(RowOffSet(i)-2)) and (Y<=(RowOffSet(i)+2)) then
                begin
                  Cursor:=crVSplit;
                  FVSizing:=True;
                  FRowResize:=i;
                  Exit;
                end
              else
                begin
                  Cursor:=crDefault;
                  FHSizing:=False;
                  FVSizing:=False;
                end;
            end;
          for i:=1 to Pred(FCol) do
            begin
              if (X>=(ColOffSet(i)-2)) and (X<=(ColOffSet(i)+2)) then
                begin
                  Cursor:=crHSplit;
                  FHSizing:=True;
                  FColResize:=i;
                  Break;
                end
              else
                begin
                  Cursor:=crDefault;
                  FHSizing:=False;
                  FVSizing:=False;
                end;
             end;
        end;
    end;
  inherited MouseMove(Shift, X, Y);
end;

procedure TTblShape.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
var i, RowHeight:Integer;
begin
  if FCanResize then
    begin
     if FHSizing then
       begin
         FColSizes.Items[FColResize-1]:=FColSizes.Items[FColResize-1]+(X-FDownX);
         FColSizes.Items[FColResize]:=FColSizes.Items[FColResize]-(X-FDownX);
         Repaint;
       end;
     if FVSizing then
       begin
         if FRowResize=1 then
           begin
             FRowSizes.Items[FRowResize-1]:=FRowSizes.Items[FRowResize-1]+(Y-FDownY);
             RowHeight:=(Height-FRowSizes.Items[FRowResize-1]) div (FRow-1);
             for i:=1 to FRow-1 do FRowSizes.Items[i]:=RowHeight;
           end
         else
           begin
             FRowSizes.Items[FRowResize-1]:=FRowSizes.Items[FRowResize-1]+(Y-FDownY);
             FRowSizes.Items[FRowResize]:=FRowSizes.Items[FRowResize]-(Y-FDownY);
           end;
         Repaint;
       end;
    end;
  inherited MouseUp(Button,Shift,X,Y);
end;

procedure TTblShape.Paint;
var i:Integer;

   procedure DrawLine(X,Y,Len,Vert:Integer);
   begin
     with Canvas do
       begin
         if Vert=0 then
           begin
             MoveTo(X, Y);
             LineTo(Len,Y);
           end
         else
           begin
             MoveTo(X, Y);
             LineTo(X,Len);
           end;
       end;
   end;

begin
  Canvas.Pen:=Pen;
  Canvas.Brush:=Brush;
  if FTransparent then
    Canvas.FillRect(Rect(0,0,Width,Height));
  if tfTop in FTabFrame then
    if Pen.Width>1 then
      DrawLine(0,Pen.Width div 2,Width,0)
    else DrawLine(0,0,Width,0);
  if tfLeft in FTabFrame then
    if Pen.Width>1 then
      DrawLine(Pen.Width div 2,0,Height,1)
    else DrawLine(0,0,Height,1);
  if tfBottom in FTabFrame then
    DrawLine(0,Height-1,Width-1,0);
  if tfRight in FTabFrame then
    DrawLine(Width-1,0,Height-1,1);
  for i:=1 to Pred(Row) do
    DrawLine(0, RowOffSet(i), Width, 0);
  for i:=1 to Pred(Col) do
    DrawLine(ColOffSet(i),0, Height, 1)
end;

function TTblShape.RowOffSet(Index: Integer): Integer;
var i:integer;
begin
  Result:=0;
  for i:=0 to pred(Index) do
    Result:=Result+FRowSizes.Items[i];
end;

procedure TTblShape.SetCanResize(const Value: Boolean);
begin
  FCanResize := Value;
end;

procedure TTblShape.SetCol(const Value: Integer);
begin
  if FCol<>Value then
    FCol := Value;
  CalcSplitSize;
  Invalidate;
end;

procedure TTblShape.SetColSizes(const Value: TSizeList);
begin
  FColSizes.AddSizes(Value);
  if FColSizes.Count<>FCol then
    FCol:=FColSizes.Count;
  Width:=FColSizes.SumTotal;
  Invalidate;
end;

procedure TTblShape.SetRow(const Value: Integer);
begin
  if FRow<>Value then
    FRow := Value;
  CalcRowSize;
  Invalidate;
end;

procedure TTblShape.SetRowSizes(const Value: TSizeList);
begin
  FRowSizes.AddSizes(Value);
  if FRowSizes.Count<>FRow then
    FRow:=FRowSizes.Count;
  Height:=FRowSizes.SumTotal;
  Invalidate;
end;

procedure TTblShape.SetTabFrame(const Value: TTabFrame);
begin
  FTabFrame := Value;
  Invalidate;
end;

procedure TTblShape.SetTransparent(const Value: Boolean);
begin
  FTransparent := Value;
  Invalidate;
end;

end.
